from typing import Dict, Iterable, Optional, Set, List, Tuple
from functools import total_ordering

from adaptor.evaluators.token_classification import TokenClassificationEvaluator
from adaptor.utils import AdaptationDataset
from more_itertools import zip_equal
import torch
from transformers import PreTrainedTokenizer

from ..config import CONFIG as _CONFIG


CategoryName = str
CategoryNames = Iterable[CategoryName]
GroupName = str
GroupNames = Iterable[GroupName]
Category = int
Group = Set[CategoryName]
CategoryMap = Dict[CategoryName, Category]
GroupMap = Dict[GroupName, Group]

FScore = float


@total_ordering
class AggregateMeanFScoreEvaluator(TokenClassificationEvaluator):
    CONFIG = _CONFIG['recognition.AggregateMeanFScoreEvaluator']
    DEFAULT_GROUP_NAMES = tuple(CONFIG['default_group_names'].split('+'))
    BETA = CONFIG.getfloat('f_score_beta')
    GROUPS: GroupMap = {
        'PER': {'B-PER', 'I-PER', 'B-ORG', 'I-ORG'},
        'LOC': {'B-LOC', 'I-LOC', 'B-ORG', 'I-ORG'},
        'O': {'O', 'B-MISC', 'I-MISC'},
    }

    def __init__(self, category_map: CategoryMap, group_names: Optional[GroupNames],
                 *args, **kwargs):
        self.category_map = category_map
        self.group_names = self.DEFAULT_GROUP_NAMES if group_names is None else tuple(group_names)
        super().__init__(*args, **kwargs)

    def __call__(self, model: torch.nn.Module, tokenizer: PreTrainedTokenizer,
                 dataset: AdaptationDataset, ignored_index: Category = -100) -> FScore:
        expected_labels, actual_labels = self._collect_token_predictions(model, dataset)

        mean_f_score, total_number_of_samples = 0, 0
        for group_name in self.group_names:
            number_of_samples, f_score = self.get_f_score(
                self.GROUPS[group_name], expected_labels, actual_labels, ignored_index)
            mean_f_score += number_of_samples * f_score
            total_number_of_samples += number_of_samples
        if total_number_of_samples > 0:
            mean_f_score /= total_number_of_samples

        return mean_f_score

    def get_f_score(self, group: Group, expected_labels: List[Category],
                    actual_labels: List[Category], ignored_index: Category) -> Tuple[int, FScore]:
        expected_categories: Set[Category] = {
            self.category_map[category]
            for category
            in group
            if category in self.category_map
        }

        true_positives, false_positives, false_negatives = 0, 0, 0
        for expected_label, actual_label in zip_equal(expected_labels, actual_labels):
            if expected_label == ignored_index:
                continue
            if expected_label in expected_categories and actual_label in expected_categories:
                true_positives += 1
            elif expected_label not in expected_categories and actual_label in expected_categories:
                false_positives += 1
            elif expected_label in expected_categories and actual_label not in expected_categories:
                false_negatives += 1

        number_of_samples = true_positives + false_positives + false_negatives

        numerator = (1.0 + self.BETA**2) * true_positives
        denominator = numerator + self.BETA**2 * false_negatives + false_positives
        f_score = numerator / denominator if numerator > 0 else 0.0

        return number_of_samples, f_score

    def __hash__(self):
        return hash(self.group_names)

    def __eq__(self, other) -> bool:
        if isinstance(other, AggregateMeanFScoreEvaluator):
            return self.group_names == other.group_names
        return NotImplemented

    def __lt__(self, other) -> bool:
        if isinstance(other, AggregateMeanFScoreEvaluator):
            return self.group_names < other.group_names
        return NotImplemented

    def __repr__(self) -> str:
        return '+'.join(self.group_names)

    def __str__(self) -> str:
        return repr(self)
